#!/usr/bin/perl
# $Id: Hypersrc.pl,v 1.101 2001/11/07 05:51:37 jimb Exp $
#
# Author   : Jim Brooks <jim@jimbrooks.org>
# Synopsis : Wrapper to automate running hypersrc.
# Usage    : See "Hypersrc.pl -help" and "hypersrc -help" for option list.
# Overview : 1. "find" is called to build a list of src files
#               (unless -srcs/-dirs is passed).
#            2. ctags.pl (which wraps/augments Exuberant ctags) is called
#               to produce lines of tags.
#            3. The tags lines are piped to "hypersrc".
# Notes    : This script is sed-edited by "make install".
# -----------------------------------------------------------------------------

use English;
use Cwd;
use File::Basename;

# =============================================================================
# ==== BEGIN CONFIGURATION ====================================================
#
# Generally, only the paths really need to be configured.
# The find/verbosity/temp/etc settings can be left as-is if you're in a hurry.


# Path to hypersrc/ctags.pl/Exuberant ctags (don't use ~ in Perl scripts).
# Red Hat 6.0 has Exuberant ctags at /usr/bin.
#
$hypersrc = "/usr/local/bin/hypersrc";
$ctags_pl = "/usr/local/bin/ctags.pl";
$ctags    = "/usr/local/bin/exctags";

# These are custom hypersrc-only args.
#
# For lo-res screen:
#$argsHypersrcCustom = "-tab-width 4  -geom 80 160 940 600 -TagModuleTypeLine -column-widths 140 90 70 20  -max-tags 100000 -font-list \"6x12\"";
#
# For med-res screen:
#$argsHypersrcCustom = "-tab-width 4  -geom 24 80 1000 680 -TagModuleTypeLine -column-widths 110 70 46 20  -max-tags 100000 -font-list \"6x12\"";
#
# For hi-res screen:
#$argsHypersrcCustom = "-tab-width 4  -geom 24 100 1200 720 -TagModuleTypeLine -column-widths 140 90 56 30  -max-tags 100000 -font-list \"6x12\"";

# Default amount of subdir levels 'find' will descend.
#
$descend = 5;

# Verbosity.
#
$verbose = 1;

# Set a short or long window title (either based on CWD).
#
$shortWindowTitleFlag = 1;

# Finally, for customization, include an optional inline Perl file.
# The inline file could override any previous state.
#
$optional_custom_file = `cat ~/.hypersrcpl 2>/dev/null`;
eval $optional_custom_file;

# ==== END CONFIGURATION ======================================================
# =============================================================================

print "\n";

# Variables.
#
@ext                  = ( );
@extTmp               = ( );
@argsHypersrc         = ( );
$ignoreNextArg        = 0;
$buildSrcListFromArgs = 0;
$buildDirListFromArgs = 0;
$srcsOption           = 0;		# true if "-srcs" passed (or implied)
$srcString            = "";		# string of lines of filenames
@srcs                 = ( );	# list   of src files
@dirs                 = ( );	# list of      dirs as passed by "-dirs" cmd arg
$ctags_pl_args        = "";
$Hypersrc_pl          = basename( $PROGRAM_NAME );
$ignoreCase           = 0;
$titleLenMax		  = 80;		# max len of chars in title string
$nowait               = 0;
$gdb				  = 0;

# Some args (eg -q) must processed very early.
# WARNING: such args will be seen twice!
#
@preArgs = ( );
foreach ( @ARGV )
{
   if ( $_ eq "-q"  ||  $_ eq "--quiet" )
   {
      @preArgs = ( @preArgs, $_ );
   }
}

# Merge default Hypersrc.pl args with passed args.
#
@argsMerged = ( @argsDefault, @ARGV );
@argsMerged = ( @preArgs, @argsMerged );

# Parse args passed to Hypersrc.pl (some may be intended for hypersrc).
#
# Filter-out args that only Hypersrc.pl understands
# by building a second arg list that Hypersrc.pl doesn't understand,
# which is assumed to be args for hypersrc.
#
$argIdx				= -1;
$argIdxPrevImplSrc	= -999;		# idx of previous implied src file
foreach ( @argsMerged )
{
   ++$argIdx;

   # Should this arg be ignored?
   #
   if ( $ignoreNextArg )
   {
      $ignoreNextArg = 0;
      next;
   }

   # Should this arg be added to the list of src files?
   #
   if ( $buildSrcListFromArgs )
   {
      # Append this arg only if it does not have a "-" prefix
      # (such args are assumed to be more options).
      #
      if ( substr( $_, 0, 1 ) ne "-" )
      {
         # Add this arg to the list of src files.
         # Go to next arg.
         #
         $srcString = $srcString . $_ . "\n";
         @srcs = ( @srcs, $_ );
         if ( $argIdx != $#argsMerged )
            { Msg( "$_ " ); }
         else
            { Msg( "$_ \n" ); }
         next;
      }

      # This arg is prefixed with "-".
      # Turn off the building of the src list.
      #
      $buildSrcListFromArgs = 0;
      Msg( "\n" );
   }

   # Should this arg be added to the list of dirs?
   #
   if ( $buildDirListFromArgs )
   {
      # Append this arg only if it does not have a "-" prefix.
      # (such args are assumed to be more options).
      #
      if ( substr( $_, 0, 1 ) ne "-" )
      {
         # Add this arg to the list of dirs.
         # Go to next arg.
         #
         @dirs = ( @dirs, "$_" );
         if ( $argIdx != $#argsMerged )
            { Msg( "$_ " ); }
         else
            { Msg( "$_ \n" ); }
         next;
      }

      # This arg is prefixed with "-".
      # Turn off the building of the dir list.
      #
      $buildDirListFromArgs = 0;
      Msg( "\n" );
   }

   if (    $_ eq "--help"
        || $_ eq "-help"
        || $_ eq "-h" )
   {
      print( "Usage:                                                                \n" );
      print( "                                                                      \n" );
      print( "   Hypersrc.pl <options> [file(s)] [dir(s)]                           \n" );
      print( "                                                                      \n" );
      print( "   (See also hypersrc -h)                                             \n" );
      print( "                                                                      \n" );
      print( "Synopsis:                                                             \n" );
      print( "                                                                      \n" );
      print( "   Hypersrc.pl is a front-end script to run hypersrc.                 \n" );
      print( "                                                                      \n" );
      print( "Default options:                                                      \n" );
      print( "                                                                      \n" );
      print( "   If no options are passed, then Hypersrc.pl will                    \n" );
      print( "   default to browsing C src files from the CWD                       \n" );
      print( "   (and will descend down a few subdir levels).                       \n" );
      print( "                                                                      \n" );
      print( "File/dir args:                                                        \n" );
      print( "                                                                      \n" );
      print( "   Pass the name(s) of src file(s) that Exuberant ctags supports.     \n" );
      print( "                                                                      \n" );
      print( "   If the name(s) of a dir(s) is passed, then this script will        \n" );
      print( "   search for src files of the specified programming langauges(s)     \n" );
      print( "   (default is C).                                                    \n" );
      print( "                                                                      \n" );
      print( "Options affecting how this script will search for files:              \n" );
      print( "                                                                      \n" );
      print( "   -descend x                                                         \n" );
      print( "        How many subdirs to descend when searching for src files.     \n" );
      print( "        (Pass 1 to descend down to one subdir level.)                 \n" );
      print( "                                                                      \n" );
      print( "   -case  or  -no-case                                                \n" );
      print( "        Whether to regard or ignore filename case, resp.              \n" );
      print( "        Default is -case (regard filename case).                      \n" );
      print( "        (-no-case only works with GNU find.)                          \n" );
      print( "                                                                      \n" );
      print( "   -lang  c|c++|perl|asm|m4|forth                                     \n" );
      print( "        Which kind of src files this script will search for.          \n" );
      print( "        Only one language can be paired with -lang,                   \n" );
      print( "        but -lang can be passed multiple times.                       \n" );
      print( "        (Note files of other languages can still be passed manually). \n" );
      print( "                                                                      \n" );
      print( "   -ext  <.filenameExtension>                                         \n" );
      print( "        If -lang does not support a particular file extension,        \n" );
      print( "        then use -ext to specify it.  Eg: \'-ext .f77\'               \n" );
      print( "        Only one extension can be paired with one -ext,               \n" );
      print( "        but -ext can be passed multiple times.                        \n" );
      print( "                                                                      \n" );
      print( "Other options:                                                        \n" );
      print( "                                                                      \n" );
      print( "   -q                                                                 \n" );
      print( "        Script runs quietly.                                          \n" );
      print( "                                                                      \n" );
      print( "   -proto                                                             \n" );
      print( "        Tag function prototypes (disabled by default).                \n" );
      print( "                                                                      \n" );
      print( "   -class-qual                                                        \n" );
      print( "        Fully qualify C++ class members.                              \n" );
      print( "        (eg show 'class::member' in tag name)                         \n" );
      print( "                                                                      \n" );
      print( "   -no-wait                                                           \n" );
      print( "        Hypersrc.pl will exit without waiting for hypersrc to exit.   \n" );
      print( "        Conserves 1M of memory, but Ctrl+C won't stop hypersrc.       \n" );
      print( "                                                                      \n" );
      print( "hypersrc-specific options:                                            \n" );
      print( "                                                                      \n" );
      print( "   Options not handled by Hypersrc.pl will be passed thru to hypersrc.\n" );
      print( "                                                                      \n" );
      print( "Examples (that descend down to a few subdirs):                        \n" );
      print( "                                                                      \n" );
      print( "   Hypersrc.pl              (defaults to C src files: .c/.h)          \n" );
      print( "   Hypersrc.pl  -lang perl  (browse Perl src files)                   \n" );
      print( "   Hypersrc.pl  -ext .cc    (browse only C++ src files ending with .cc)\n" );
      print( "   Hypersrc.pl  myprog/src  (browse C src files descending from this subdir)\n" );
      print( "                                                                      \n" );
      print( "Examples (non-descending):                                            \n" );
      print( "                                                                      \n" );
      print( "   Hypersrc.pl  *.f77       (browse Fortran src files)                \n" );
      print( "                                                                      \n" );
      print( "Jim Brooks jim\@jimbrooks.org                                         \n" );
      print( "Updates available at: http://www.jimbrooks.org/web/hypersrc/hypersrc.html \n" );
      print( "\n" );

      #system "${hypersrc} -help | more";

      exit 0;
   }

   if ( $_ eq "-srcs"  ||  $_ eq "-src" )
   {
      # All args that follow, without a "-" prefix,
      # are assumed to be names of src files.
      #
      # Set a flag so that the next arg iterations will build
      # a list of src files, until an arg is found with a "-" prefix.
      #
      $buildSrcListFromArgs = 1;

      # Preamble, later will print each parsed src file.
      #
      MaybePrintNewline();
      Msg( "Will browse src files: " );

      # Set a flag (which stays set) to remember that "-srcs" was specified.
      # This flag will later cause calling "find" to be bypassed.
      #
      $srcsOption = 1;

      # If this isn't a C/C++ file, then pass (dangerously) -accept-all-tags
      # to hypersrc.
      #
      CheckFilename( $_ );

      next;
   }

   if ( $_ eq "-dirs"  ||  $_ eq "-dir" )
   {
      # All args that follow, without a "-" prefix,
      # are assumed to be names of dirs (that contain src files).
      #
      # Set a flag so that the next arg iterations will build
      # a list of dirs, until an arg is found with a "-" prefix.
      #
      $buildDirListFromArgs = 1;

      # Preamble, later will print each parsed dir.
      #
      MaybePrintNewline();
      Msg( "Will descend dir: " );

      next;
   }

   if ( $_ eq "-descend"  ||  $_ eq "-maxdepth" )
   {
      MaybePrintNewline();
      $x = $argsMerged[$argIdx+1];
      Msg( "Will descend no more than $x subdir levels. \n" );

      $descend = $argsMerged[$argIdx+1] + 1;

      $ignoreNextArg = 1;
      next;
   }

   if ( $_ eq "-class-qual" )
   {
      # This arg will be passed to, and handled by, ctags.pl.
      #
      $ctags_pl_args = $ctags_pl_args . " -class-qual ";
      next;
   }

   if ( $_ eq "-proto" )
   {
      # This arg will be passed to, and handled by, ctags.pl.
      #
      $ctags_pl_args = $ctags_pl_args . " -proto ";
      next;
   }

   if ( $_ eq "-ext" )
   {
      # The arg after "-ext" is assumed to be a file extension.
      #
      $ext = $argsMerged[$argIdx+1];
      if ( substr( $ext, 0, 1 ) ne '.' ) { $ext = "." . $ext; }		# prepend "." to ext if omitted
      MaybePrintNewline();
      Msg( "Will browse files with extension \'${ext}\'.\n" );
      @extTmp = ( @extTmp, $ext );

      ${argsHypersrcCustom} = ${argsHypersrcCustom} . " -accept-all-tags ";

      $ignoreNextArg = 1;
      next;
   }

   if ( $_ eq "-lang"  &&  $argsMerged[$argIdx+1] eq "c" )
   {
      MaybePrintNewline();
      Msg( "Will browse C language src files.\n" );
      @extTmp = ( @extTmp, ".c" );	   	# -- don't add .C/.H (remember there is -no-case) --
      @extTmp = ( @extTmp, ".h" );

      $ignoreNextArg = 1;
      next;
   }

   if ( $_ eq "-lang"  &&  $argsMerged[$argIdx+1] eq "c++" )
   {
      MaybePrintNewline();
      Msg( "Will browse C++ language src files.\n" );
      @extTmp = ( @extTmp, ".cc" );     # -- don't add .CC (remember there is -no-case) --
      @extTmp = ( @extTmp, ".cxx" );
      @extTmp = ( @extTmp, ".cpp" );	# a dumb ext for C++ src
      if ( ! $ignoreCase )
    { @extTmp = ( @extTmp, ".C" ); }	# another dumb one
      @extTmp = ( @extTmp, ".h" );
      @extTmp = ( @extTmp, ".hh" );
      @extTmp = ( @extTmp, ".hxx" );
      @extTmp = ( @extTmp, ".inl" );	# inline functions (specific to C++ ?)

      $ignoreNextArg = 1;
      next;
   }

   if ( $_ eq "-lang"  &&  $argsMerged[$argIdx+1] eq "asm" )
   {
      MaybePrintNewline();
      Msg( "Will browse assembly language src files.\n" );
      @extTmp = ( @extTmp, ".s" );
      if ( ! $ignoreCase )
    { @extTmp = ( @extTmp, ".S" );  }	# asm file that needs cpp preprocessing
      @extTmp = ( @extTmp, ".sm4" );	# asm file that needs M4 preprocesing
      @extTmp = ( @extTmp, ".asm" );	# -- don't add .ASM (remember there is -no-case) --
      @extTmp = ( @extTmp, ".inc" );

      ${argsHypersrcCustom} = ${argsHypersrcCustom} . " -accept-all-tags ";

      $ignoreNextArg = 1;
      next;
   }

   if ( $_ eq "-lang"  &&  $argsMerged[$argIdx+1] eq "perl" )
   {
      MaybePrintNewline();
      Msg( "Will browse Perl script src files.\n" );
      @extTmp = ( @extTmp, ".pl" );		# -- don't add .PL (remember there is -no-case) --

      ${argsHypersrcCustom} = ${argsHypersrcCustom} . " -accept-all-tags ";

      $ignoreNextArg = 1;
      next;
   }

   if ( $_ eq "-lang"  &&  $argsMerged[$argIdx+1] eq "m4" )
   {
      MaybePrintNewline();
      Msg( "Will browse M4 macro src files.\n" );
      @extTmp = ( @extTmp, ".m4" );	# -- don't add .M4 (remember there is -no-case) --

      ${argsHypersrcCustom} = ${argsHypersrcCustom} . " -accept-all-tags ";

      $ignoreNextArg = 1;
      next;
   }

   if ( $_ eq "-lang"  &&  $argsMerged[$argIdx+1] eq "forth" )
   {
      MaybePrintNewline();
      Msg( "Will browse FORTH src files.\n" );
      @extTmp = ( @extTmp, ".4th" );	# -- don't add .4TH (remember there is -no-case) --
      @extTmp = ( @extTmp, ".fth" );

      ${argsHypersrcCustom} = ${argsHypersrcCustom} . " -accept-all-tags ";

      $ignoreNextArg = 1;
      next;
   }

   if ( $_ eq "-gdb" )
   {
      $gdb = 1;
      next;
   }

   if ( $_ eq "-q"  ||  $_ eq "--quiet" )
   {
      $verbose = 0;
      next;
   }

   # This arg wasn't recognized.
   # Is it a file?
   #
   if ( -f $_ )
   {
      # For readability, remove "./" prefix (if "find" was used).
      #
      $fname = $_;
      if ( substr( $fname, 0, 2 ) eq "./" )
      { $fname = substr( $fname, 2 ); } 

      # Add this arg to the list of src files.
      # Go to next arg.
      #
      $srcString = $srcString . $fname . "\n";
      @srcs = ( @srcs, $fname );
      if ( $argIdxPrevImplSrc + 1 != $argIdx )
      { Msg( "Src file(s) to browse: " ); }
      Msg( "$fname " );
      $argIdxPrevImplSrc = $argIdx;

      # The -src option is implied.
      # -src inhibits the default behavior finding C files.
      #
      $srcsOption = 1;

      # If this isn't a C/C++ file, then pass (dangerously) -accept-all-tags
      # to hypersrc.
      #
      CheckFilename( $fname );

      next;
   }

   # This arg wasn't recognized.
   # Is it a directory?
   #
   if ( -d $_ )
   {
      # Add this arg to the list of dirs.
      # Go to next arg.
      #
      MaybePrintNewline();
      @dirs = ( @dirs, "$_" );
      $dir = $_;
      if ( substr( $dir, length($dir)-1, 1 ) ne "/" ) { $dir = $dir . "/"; }
      Msg( "Will descend dir: ${dir} \n" );
      next;
   }

   # A file-glob can reach this point unexpanded if no files exist that match it.
   #
   if ( substr($_, 0, 2) eq "*."  ||  substr($_, 0, 1) eq "*" )
   {
      MaybePrintNewline();
      print "(no files match '$_') \n";
      next;
   }

   # -case ?
   #
   if ( $_ eq "-case" )
   {
      $ignoreCase = 0;
      next;
   }

   # -no-case ?
   #
   if ( $_ eq "-no-case"  ||  $_ eq "-nocase" )
   {
      $ignoreCase = 1;
      next;
   }

   # -no-wait ?
   #
   if ( $_ eq "-no-wait"  ||  $_ eq "-nowait" )
   {
      $nowait = 1;
      next;
   }

   # -version ?
   #
   if ( $_ eq "-version" || $_ eq "--version")
   {
      @argsHypersrc = ( @argsHypersrc, $_ );
      goto run_hypersrc_now;
   }

   # This arg is not understood by Hypersrc.pl.
   # Add it to the arg list that is assumed to be for hypersrc.
   #
   @argsHypersrc = ( @argsHypersrc, $_ );
}

MaybePrintNewline(1);

# Check that the paths are correct.
#
CheckPaths();

# Default to CWD if no files/dirs were passed.
#
if ( ! scalar(@dirs)  &&  ! $srcsOption )
{ @dirs = ( "." ); }

# Ensure that no file extension is duplicated.
# Output will be in @ext.
#
FilterDupFileExt();

# If no file extensions were specified and there is a dir list,
# then default to C src files.
#
if ( ! scalar(@ext)  &&  scalar(@dirs) )
{
   Msg( "Defaulting to C src files.\n" );
   @ext = ( ".c", ".h" );
}

# Show what file extensions will be browsed.
#
if ( scalar(@ext) ) { Msg( "Src file extensions to browse: @ext\n" ); }

# Build a list of src files based on dir list.
# Removing the leading "./" that is prepended by "find".
# Exclude "RCS/".
#
if ( scalar(@dirs) )
{
   Msg( "Building list of src files (patience)...\n" );
   foreach $item (@ext)
   {
      foreach $dir (@dirs)
      {
         # Invoke "find" to print each filename on a separate line,
         # so that ctags.pl can subsequently read filenames by line via stdin.
         #
         if ( $ignoreCase )
         { $srcString = $srcString.`find $dir -maxdepth ${descend} -iname "*$item" 2>/dev/null `; }
         else
         { $srcString = $srcString.`find $dir -maxdepth ${descend}  -name "*$item" 2>/dev/null `; }
      }
   }
   $srcString =~ s/^\.\///;         # (first)    remove "./" but not "../"
   $srcString =~ s/\n\.\//\n/g;     # (the rest) remove "./" but not "../"
#  $srcString =~ s/\bRCS.*\b//;
   $srcString =~ s/.*RCS.*\n//g;
}

# Ensure the list of src files is not empty.
#
if ( $srcString eq "" )
{
   print( "## No src files were found nor specified.  Quitting. \n" );
   print( "## Perhaps try passing -no-case to find MSDOG src files. \n" );
   exit 1;
}

# Title the main window based on list of files/dirs.
#
$title = "";
if ( scalar(@dirs) )
{
   BuildTitleFromList( "dirs" );
   if ( $srcsOption  &&  length($title)+2 < $titleLenMax )
   { $title = $title . ", "; }
}
if ( $srcsOption )
{
   BuildTitleFromList( "srcs" );
}


###################
# Debug: dump ctags output to a file
#
if ( 0 )
{
   Msg( "DEBUG: dumping tags to /tmp/tags.tmp \n" );
   open( PIPE, "| ${ctags_pl}  -ctags-path ${ctags}  -stdin  ${ctags_pl_args}  -enable-non-tag  > /tmp/tags.tmp" );
   print PIPE "$srcString";
   close( PIPE );
   exit 0;
}
###################

########################
# Run hypersrc under gdb
if ( $gdb )
{
   # Create gdb command file (for hypersrc args).
   #
   $cmdFile = "/tmp/hypersrc_gdb_${PID}.tmp";
   open( CMD_FILE, ">$cmdFile" );
   print CMD_FILE "run -stdin ${argsHypersrcCustom} @argsHypersrc -title 'hypersrc (${title})' \n";
   print CMD_FILE "q \n";
   close( CMD_FILE );

   Msg( "Piping ctags into hypersrc (please wait)... \n" );
   open( PIPE, "| ${ctags_pl}  -ctags-path ${ctags}  -stdin  ${ctags_pl_args}  -enable-non-tag  |  gdb -x $cmdFile ${hypersrc} " );
   print PIPE "$srcString";
   close( PIPE );

   system "rm -f $cmdFile";

   exit 0;
}
########################

# Start the pipeline.
# Begin by piping the list of filenames to ctags.pl.
# (Piping is necessary because list might be too long to be passed via cmd-line).
#
Msg( "Piping ctags into hypersrc (please wait)... \n" );
run_hypersrc_now:
open( PIPE, "| ${ctags_pl}  -ctags-path ${ctags}  -stdin  ${ctags_pl_args}  -enable-non-tag  |  ${hypersrc} -stdin ${argsHypersrcCustom} @argsHypersrc -title 'hypersrc (${title})'" );
print PIPE "$srcString";
if ( ! $nowait )
{ close( PIPE ); }
else
{ print "(hypersrc was started)\n"; }

# Done.
#
exit 0;

# =============================================================================
# subroutines
# =============================================================================

sub Msg
{
   # Only print a message if $verbose.
   #
   if ( $verbose ) { print "$_[0]"; }
}

sub IsHypersrcExecutable
{
   # Return true if $hypersrc points to an executable binary.
   # Ignore hypersrc dir (dirs are also executable).
   #
   stat( "$hypersrc" );
   if ( -x _  &&  ! -d _ ) { return 1; }
   return 0;
}

sub TestInvokingHypersrc
{
   # Return true if hypersrc can be invoked.
   #
   $_ = `$hypersrc -h`;
   return defined($_)  &&  ($_ ne "");
}

sub FilterDupFileExt
{
   # @extTmp holds the original file extensions.
   # @ext    is the output which holds only unique file extensions.

   %seen = ();
   @ext  = ();
   foreach $item (@extTmp)
   {
      unless ($seen{$item})
      {
         # Never seen before.
         #
         $seen{$item} = 1;
         push(@ext, $item);
      }
   }
}

sub CheckPaths
{
   # Ensure $hypersrc points to an executable.
   # If not, try to correct $hypersrc.
   #
   if ( ! IsHypersrcExecutable() )
   {
#      print( "## Configured path to hypersrc is wrong, defaulting to PATH. \n" );

      # Remember original paths.
      #
      $hypersrc_orig = $hypersrc;
      $ctags_pl_orig = $ctags_pl;

      # Try to locate hypersrc using PATH.
      #
      $hypersrc  =  "hypersrc";		# <-- extra spaces to fool Makefile install

      if ( ! IsHypersrcExecutable() )
      {
         # Perhaps Hypersrc.pl was invoked after "make install" was done?
         #
         $hypersrc = $Hypersrc_pl;
         $hypersrc =~ s/${Hypersrc_pl}/hypersrc/;

         $ctags_pl = $Hypersrc_pl;
         $ctags_pl =~ s/${Hypersrc_pl}/ctags.pl/;
      }
      if ( ! IsHypersrcExecutable() )
      {
         # Perhaps Hypersrc.pl was invoked without "make install"?
         #
         $hypersrc = $Hypersrc_pl;
         $hypersrc =~ s/${Hypersrc_pl}/out\/hypersrc/;

         $ctags_pl = $Hypersrc_pl;
         $ctags_pl =~ s/${Hypersrc_pl}/ctags.pl/;
      }
      if ( ! IsHypersrcExecutable() )
      {
         # Perhaps Hypersrc.pl is a symlink?
         #
         $hypersrc = readlink "$PROGRAM_NAME";
         chomp $hypersrc;
         $hypersrc =~ s/${Hypersrc_pl}/out\/hypersrc/;

         $ctags_pl = readlink "$PROGRAM_NAME";
         chomp $ctags_pl;
         $ctags_pl =~ s/${Hypersrc_pl}/ctags.pl/;
      }
      if ( ! IsHypersrcExecutable() )
      {
         # Ugghhh, the programs appear to not be executable and/or outside PATH.
         # Charge forward with invocation test, in case this Perl code here is mistaken.
         #
         #print( "## This script\'s paths seem misconfigured, proceeding anyway... ##\n" );
         $hypersrc = $hypersrc_orig;
         $ctags_pl = $ctags_pl_orig;
      }
   }

   # Ensure hypersrc can be invoked.
   #
   if ( ! TestInvokingHypersrc() )
   {
      print( "## Failed to invoke hypersrc. \n" );
      print( "## Try 'make install' (or configure this script). \n" );
      exit 1;
   }

   # Ensure there is a recent version of Exuberant ctags.
   #
   $_ = `$ctags --version 2>/dev/null | grep Exuberant`;
   if ( "" eq $_  ||  ! defined($_) )
   {
      # Perhaps this is not Red Hat 6.0 but Exuberant ctags is at /usr/local/bin?
      #
      $ctags ="/usr/local/bin/ctags";
   }
   $_ = `$ctags --version 2>/dev/null | grep Exuberant`;
   if ( "" eq $_  ||  ! defined($_) )
   {
      # Try PATH.
      #
      $ctags ="ctags";
   }
   $_ = `$ctags --version 2>/dev/null | grep Exuberant`;
   if ( "" eq $_  ||  ! defined($_) )
   {
      print( "## Script could not find a recent version of Exuberant ctags.\n" );
      print( "## Try configuring this script.\n" );
#     print( "## Exuberant ctags is available at http://fly.hiwaay.net/~darren/ctags \n" );
      print( "## Exuberant ctags is available at http://ctags.sourceforge.net        \n" );
      exit 1;
   }
}

# Append to title string from src or dir list.
# Parms: $1 = "dirs" or "srcs"
#        $title (this subroutine appends to this string)
#
sub BuildTitleFromList
{
   my $dirs_or_srcs = $_[0];
   my @list;
   my $item;

   # Use either the file or dir list.
   #
   if    ( "$dirs_or_srcs" eq "dirs" )
   {
      @list = @dirs;

      # If the only dir is "." (no dirs were passed by user)
      # then determine the name of CWD.
      #
      if ( scalar(@list) == 1  &&  $list[0] eq "." )
      {
         if ( $shortWindowTitleFlag )
            { $list[0] = basename( cwd() ); }
         else
            { $list[0] = cwd(); }
      }
   }
   elsif ( "$dirs_or_srcs" eq "srcs" )
   {
      @list = @srcs;
   }
   else
   { die; }

   # Accumulate a limited amount of items into the title string.
   #
   for ( 0 .. $#list )
   {
      if ( length($title) > $titleLenMax ) { last; }

      if ( $shortWindowTitleFlag )
      {
         # basename() will be confused by a superfluous trailing "/".
         #
         $item = $list[$_];
         if ( substr( $item, length($item)-1, 1 ) eq "/" )
         { chop $item; }

         if ( $_ > 0 ) { $title = $title . ", "; }
         $title = $title . basename( $item );

         if ( "$dirs_or_srcs" eq "dirs" )
         { $title = $title . "/"; }
      }
      else
      {
         $title = "${srcs_or_dirs}: $list[0], [...]";
         last;
      }
   }

   # Add ellipsis if there were too many items.
   #
   if ( length($title) > $titleLenMax ) { $title = $title . ", [...]"; }
}

# Check if a filename is a C/C++ src file.
# If not, then prepare to pass -accept-all-tags to hypersrc.
#
sub CheckFilename
{
   my $filename = $_[0];
   my $ext;
   my $match;

   $ext = substr( $_, rindex($_,".") );
   $match  = index( $ext, "-.c.cc.cpp.cxx.h.hh.hxx.inl" );
   $match2 = index( $ext, "-.C.CC.CPP.CXX.H.HH.HXX.INL" );	# in case of MSDOG files
   if ( $match == -1  &&  $match2 == -1 )
   { ${argsHypersrcCustom} = ${argsHypersrcCustom}." -accept-all-tags "; }
}

# Print a newline depending on how previous arg was printed.
# Pass nothing if within the arg loop, pass 1 after leaving the loop.
#
sub MaybePrintNewline
{
   my $inTheLoop = 1;

   # Still within the arg loop?
   #
   if ( defined($_[0])  &&  $_[0] == 1 )
   { $inTheLoop = 0; }

   # A newline wasn't printed if previous arg was an implied src file.
   #
   if ( $argIdxPrevImplSrc + $inTheLoop == $argIdx )
   { Msg( "\n" ); }
}
